import {configDependency} from '../core/Config.js'
import {
   Dependency, ImpType
} from '../core/Dispatcher.js'
import type {Complex, UnderlyingReal} from '../Complex/type.js'

type UnaryNumber = (a: number) => number
type BinaryNumber = (a: number, b:number) => number

declare module "./type" {
   interface NumbersImpTypes {
      add: BinaryNumber
      unaryMinus: UnaryNumber
      conj: UnaryNumber
      subtract: BinaryNumber
      multiply: BinaryNumber
      absquare: UnaryNumber
      reciprocal: UnaryNumber
      divide: BinaryNumber
      conservativeSqrt: UnaryNumber
      // Best we can do for sqrt at compile time, since actual return
      // type depends on config. Not sure how this will play out
      // when we make a number-only bundle, but at least the import type
      // above for Complex<> does not lead to any emitted JavaScript.
      sqrt: (a: number) => number | Complex<number>
   }
}

export const add: ImpType<'add', [number, number]> = (a, b) => a + b
export const unaryMinus: ImpType<'unaryMinus', [number]> = a => -a
export const conj: ImpType<'conj', [number]> = a => a
export const subtract: ImpType<'subtract', [number, number]> = (a, b) => a - b
export const multiply: ImpType<'multiply', [number, number]> = (a, b) => a * b
export const absquare: ImpType<'absquare', [number]> = a => a*a
export const reciprocal: ImpType<'reciprocal', [number]> = a => 1/a
export const divide: ImpType<'divide', [number, number]> = (a, b) => a / b

export const conservativeSqrt: ImpType<'conservativeSqrt', [number]> =
   a => isNaN(a) ? NaN : Math.sqrt(a)

export const sqrt =
   (dep: configDependency
    & Dependency<'complex', [number, number]>): ImpType<'sqrt', [number]> => {
       if (dep.config.predictable || !dep.complex) return conservativeSqrt
       return a => {
          if (isNaN(a)) return NaN
          if (a >= 0) return Math.sqrt(a)
          return dep.complex(0, Math.sqrt(unaryMinus(a)))
       }
    }
